@@ -13,7 +13,7 @@
 
  Author: Chris Marrison
 
- Date Last Updated: 20210730
+ Date Last Updated: 20210804
 
  Todo:
 
@@ -45,7 +45,7 @@
 
 ------------------------------------------------------------------------
 '''
-__version__ = '0.2.0'
+__version__ = '0.2.5'
 __author__ = 'Chris Marrison'
 __author_email__ = 'chris@infoblox.com'
 
@@ -242,14 +242,16 @@ class dhcp_encode():
         i_sizes = [ 8, 16, 32 ]
         if size in i_sizes:
             max_size = 2**size
-            no_bytes = size // 4
-            fmt = f'{{:0{no_bytes}x}}'
-            if i < max_size:
-                hex = fmt.format(i + (2**size))
+            no_octets = size // 4
+            fmt = f'{{:0{no_octets}x}}'
+            if i < 0 and abs(i) < max_size:
+                hex_str = fmt.format(i + (2**size))
+            elif i < max_size:
+                hex_str = fmt.format(i)
             else:
                 raise TypeError(f'{i} is out of range for uint{size} type')
 
-        return hex 
+        return hex_str 
 
 
     # Methods for intX and uintX
@@ -937,6 +939,14 @@ class dhcp_decode():
 
     def hex_to_array_of_ip(self, hex_string):
         '''
+        Decode array of IPv4 or IPv6 addresses to CSV string 
+
+        Parameters:
+            hex_string (str): Hex representation of an array of IPv4 or IPv6
+        
+        Returns:
+            IP Addresses in a CSV string
+        
         '''
         array_of_ip = ''
         
@@ -1022,15 +1032,15 @@ class dhcp_decode():
 
         # Assume true if not zero i.e. check all bits for non-zero
         if int(hex_string, 16) == 0:
-            text_bool = 'false'
+            text_bool = 'False'
         else:
-            text_bool = 'true'
+            text_bool = 'True'
         
         return text_bool
 
 
    # integer encodings 
-    def hex_to_int(self, hex_string, size = 8):
+    def hex_to_int(self, hex_string, size=8):
         '''
         Decode hex to signed integer of defined size
 
@@ -1061,7 +1071,7 @@ class dhcp_decode():
         return value 
 
 
-    def hex_to_uint(self, hex_string, size = 8):
+    def hex_to_uint(self, hex_string, size=8):
         '''
         Encode integer of specified size as unsigned int in hex
         Uses 2's compliment if supplied with negative number
@@ -1175,7 +1185,7 @@ class dhcp_decode():
         if data[:2] == '0b':
             base = 2
         else:
-            hex_string = hex_string.replace(':','')
+            hex_string = data.replace(':','')
 
         # Force hex encoding without 0x using base
         hex_value = '{:02x}'.format(int(data, base))
@@ -1229,7 +1239,7 @@ class dhcp_decode():
         return hex_len
 
 
-    def check_data_type(self, optcode, sub_defs=[], guess=False):
+    def check_data_type(self, optcode, sub_defs=[]):
         '''
         Get data_type for optcode from sub optino definitions
 
@@ -1246,12 +1256,40 @@ class dhcp_decode():
 
         if sub_defs:
             for d in sub_defs:
-                if optcode == d['code']:
+                if int(optcode) == int(d['code']):
                     data_type = d['type']
+                    # Check for array_of_ip 
+                    if 'array' in d.keys():
+                        if ('ip' in data_type) and d['array']:
+                            data_type = 'array_of_ip'
                     break
         
         return data_type
-                
+ 
+
+    def get_name(self, optcode, sub_defs=[]):
+        '''
+        Get data_type for optcode from sub optino definitions
+
+        Parameters:
+            optcode (int): Option code to check
+            sub_defs (list of dict): sub option definitions to cross
+                    reference
+        
+        Returns:
+            name as str
+        
+        '''
+        name = ''
+
+        if sub_defs:
+            for d in sub_defs:
+                if optcode == d['code']:
+                    name = d['name']
+                    break
+        
+        return name
+                            
 
     def guess_data_type(self, subopt, padding=False):
         '''
@@ -1299,11 +1337,16 @@ class dhcp_decode():
         return data_type
         
     
-    def decode_data(self, data, data_type='string'):
+    def decode_data(self, data, data_type='string', 
+                    padding=False,
+                    pad_bytes=1,
+                    array=False):
         '''
         '''
         decoded = ''
         if data_type in self.opt_types:
+            if 'ip' in data_type and array:
+                data_type = 'array_of_ip'
             hex_to_type = eval('self.' + 'hex_to_' + data_type)
         else:
             logging.error(f'Unsupported Option Type {data_type}')
@@ -1343,6 +1386,7 @@ class dhcp_decode():
         suboptions = []
         de_sub_opt = {}
         decoded_opts = []
+        parent = {}
         guessed = False
         hex_string = hex_string.replace(':','')
 
@@ -1350,6 +1394,8 @@ class dhcp_decode():
             parent_opt = self.hex_to_opcode(hexstring[:2])
             total_len = self.hex_to_int8(hexstring[2:4])
             hex_string = hex_string[4:]
+            parent = {'parent': parent_opt, 'total_len': total_len }
+            decoded_opts.append(parent)
 
         # Break out sub-options
         suboptions = self.hex_to_suboptions(hex_string)
@@ -1359,17 +1405,21 @@ class dhcp_decode():
             if sub_opt_defs:
                 data_type = self.check_data_type(opt['code'],
                                                  sub_defs=sub_opt_defs)
+                name = self.get_name(opt['code'], sub_defs=sub_opt_defs)
+
             else:
                 logging.debug(f'Attempting to guess option type for {opt}')
                 data_type = self.guess_data_type(opt)
                 guessed = True
+                name = 'Undefined'
             
             if data_type: 
                 value = self.decode_data(opt['data'], data_type=data_type)
             
             str_value = self.decode_data(opt['data'], data_type='string')
 
-            de_sub_opt = { 'code': opt['code'],
+            de_sub_opt = { 'name': name,
+                           'code': opt['code'],
                            'type': data_type,
                            'data_length': opt['data_length'],
                            'data': value,
@@ -1383,13 +1433,59 @@ class dhcp_decode():
 
     def output_decoded_options(self, decoded_opts=[], output='pprint'):
         '''
+        Simple output for decode_dhcp_options() data
+
+        Parameters:
+            decoded_opts (list): List of dict
+            output (str): specify format [pprint, csv, yaml]
+        
         '''
-        formats = [ 'csv', 'pprint', 'json']
-        if output in formats:
-            if output == 'pprint':
-                pprint.pprint(decoded_opts)
+        formats = [ 'csv', 'pprint', 'yaml']
+        header = ''
+        head_printed = False
+        line = ''
+        
+        if len(decoded_opts):
+            if output in formats:
+                # Output simply with pprint
+                if output == 'pprint':
+                    pprint(decoded_opts)
+                # Output to CSV
+                if output == 'csv':
+                    for item in decoded_opts: 
+                        if 'parent' in item.keys():
+                            header = 'Parent, Total Length'
+                            pprint(header)
+                            pprint(f'{item["parent"]}, ' +
+                                  f'{item["total_len"]}')
+                        elif not head_printed:
+                            header = ''
+                            for key in item.keys():
+                                header += key + ','
+                            header = header[:-1]
+
+                            pprint(header)
+                            head_printed = True
+                        else:
+                            for key in item.keys():
+                                line += repr(item[key]) + ','
+                            line += line[:-1]
+                            pprint(line)
+                            line = ''
+                # Output to normalised YAML
+                if output == 'yaml':
+                    try:
+                        y = yaml.safe_dump(decoded_opts)
+                        print(y)
+                    except:
+                        print('Could not convert to yaml')
+
+            else:
+                print(f'{output} not supported for output')
+                print(f'Suported formats include: {formats}')
+                print(decoded_opts)
         else:
-            print(decoded_opts)
+            print('No option data')
         
         return
         
@@ -1398,7 +1494,7 @@ class dhcp_decode():
         '''
         Run through encoding methods and output example results
         '''
-        encode = dhcp_encode()
+        encode = bloxone.dhcp_encode()
         test_data = [ { 'code': '1', 'type': 'string',
                         'data': 'AABBDDCCEEDD-aabbccddeeff' },
                       { 'code': '2', 'type': 'ipv4_address',
@@ -1428,17 +1524,24 @@ class dhcp_decode():
         print()
         print('Non-array tests:')
         for data_test in test_data:
-            result = encode.encode_data(data_test)
-            hex_len = self.hex_length(result)
-            print(f'Type: {data_test["type"]}: {data_test["data"]}, ' +
-                  f'Encoded: {result}, Length(hex): {hex_len}')
+            enc_str = encode.encode_data(data_test)
+            if 'array' in data_test.keys():
+                array = data_test['array']
+            else:
+                array=False
+
+            dec_str = self.decode_data(enc_str, 
+                                       data_type=data_test['type'],
+                                       array=array)
+            print(f'Type: {data_test["type"]}, Hex: {enc_str}, ' +
+                  f'Decoded: {dec_str}, Original: {data_test["data"]}')
         
         print()
         # Padding Test
-        test_data = { 'code': '99', 'type': 'string', 'data': 'AABBCCDD' }
-        result = encode.encode_data(test_data, padding=True)
-        print(f'Padding test (1 byte), type string: {test_data["data"]}' +
-              f' {result}')
+        # test_data = { 'code': '99', 'type': 'string', 'data': 'AABBCCDD' }
+        # result = encode.encode_data(test_data, padding=True)
+       #  print(f'Padding test (1 byte), type string: {test_data["data"]}' +
+        #       f' {result}')
         # Full encode test
         test_data = [ { 'code': '1', 'type': 'string',
                         'data': 'AABBDDCCEEDD-aabbccddeeff' },
@@ -1449,9 +1552,9 @@ class dhcp_decode():
                       { 'code': '4', 'type': 'boolean', 'data': True },
                       { 'code': '5', 'type': 'int8', 'data': '22' } ]
         result = encode.encode_dhcp_option(test_data)
-        print(f'Full encoding of sample: {result}')
-        decode = self.decode_dhcp_option(result)
+        print(f'Full encoding of sample Hex: {result}')
+        decode = self.decode_dhcp_option(result, sub_opt_defs=test_data)
         print(f'Decoding result:')
-        pprint.pprint(decode)
+        self.output_decoded_options(decode)
 
         return
\ No newline at end of file
